"""gustaf/spline/microstructure/tiles/__init__.py.

Interface for tools and generators creating simple microstructures.
"""

from gustaf.spline.microstructure.tiles import (
    crosstile2d,
    crosstile3d,
    inversecrosstile3d,
    tilebase,
)
from gustaf.spline.microstructure.tiles.crosstile2d import CrossTile2D
from gustaf.spline.microstructure.tiles.crosstile3d import CrossTile3D
from gustaf.spline.microstructure.tiles.inversecrosstile3d import (
    InverseCrossTile3D,
)
from gustaf.spline.microstructure.tiles.tilebase import TileBase

__all__ = [
    "tilebase",
    "crosstile3d",
    "crosstile2d",
    "inversecrosstile3d",
    "TileBase",
    "CrossTile3D",
    "CrossTile2D",
    "InverseCrossTile3D",
]



"""gustaf/gustaf/io/nutils.py.

io functions for nutils.
"""
import numpy as np

from gustaf.faces import Faces
from gustaf.io import mixd
from gustaf.io.ioutils import abs_fname, check_and_makedirs
from gustaf.volumes import Volumes


def load(fname):
    """nutils load.
    Loads a nutils (np.savez) file and returns a Gustaf Mesh.

    Parameters
    -----------
    fname: str
      The npz file needs the following keys:
      nodes, cnodes, coords, tags, btags, ptags.

    Returns
    --------
    mesh: Faces or Volumes
    """
    npz_file = np.load(fname, allow_pickle=True)
    nodes = npz_file["nodes"]
    _ = npz_file["cnodes"]
    coords = npz_file["coords"]
    _ = npz_file["tags"].item()
    btags = npz_file["btags"].item()
    _ = npz_file["ptags"].item()

    if nodes.shape[0] == 0:
        raise TypeError("Can not find nodes. Check nutils mesh description.")
    if coords.shape[0] == 0:
        raise TypeError("Can not find coords. Check nutils mesh description.")

    vertices = coords

    # connec
    simplex = True
    connec = nodes

    if vertices.shape[1] == 2:
        volume = False
    else:
        volume = True

    # reshape connec
    try:
        ncol = int(3) if simplex and not volume else int(4)
        connec = connec.reshape(-1, ncol)
        mesh = Volumes(vertices, connec) if volume else Faces(vertices, connec)
    except BaseException:
        raise RuntimeError(
            "Can not generate a mesh from the nutils input."
            "Check nutils mesh description."
        )

    mesh.BC = btags
    return mesh


def export(mesh, fname):
    """Export in Nutils format. Files are saved as np.savez().
    Supports triangle,and tetrahedron Meshes.

    Parameters
    -----------
    mesh: Faces or Volumes
    fname: str

    Returns
    --------
    None
    """

    dic = to_nutils_simplex(mesh)

    # prepare export location
    fname = abs_fname(fname)
    check_and_makedirs(fname)

    np.savez(fname, **dic)


def to_nutils_simplex(mesh):
    """Converts a Gustaf_Mesh to a Dictionary, which can be interpreted
    by ``nutils.mesh.simplex(**to_nutils_simplex(mesh))``. Only works for
    Triangles and Tetrahedrons!

    Parameters
    -----------
    mesh: Faces or Volumes

    Returns
    --------
    dic_to_nutils: dict
    """

    vertices = mesh.vertices
    faces = mesh.faces
    whatami = mesh.whatami

    # In 2D, element = face. In 3D, element = volume.
    if whatami.startswith("tri"):
        dimension = 2
        permutation = [1, 2, 0]
        elements = faces
    elif whatami.startswith("tet"):
        dimension = 3
        permutation = [2, 3, 1, 0]
        volumes = mesh.volumes
        elements = volumes
    else:
        raise TypeError("Only Triangle and Tetrahedrons are accepted.")

    dic_to_nutils = dict()

    # Sort the Node IDs for each Element.
    sort_array = np.argsort(elements, axis=1)
    elements_sorted = np.take_along_axis(elements, sort_array, axis=1)

    # Let`s get the Boundaries
    bcs = dict()
    bcs_in = mixd.make_mrng(mesh)
    bcs_in = np.ndarray.reshape(
        bcs_in, (int(len(bcs_in) / (dimension + 1)), (dimension + 1))
    )

    bound_id = np.unique(bcs_in)
    bound_id = bound_id[bound_id > 0]

    # Reorder the mrng according to nutils permutation: swap columns
    bcs_in[:, :] = bcs_in[:, permutation]

    # Let's reorder the bcs file with the sort_array
    bcs_sorted = np.take_along_axis(bcs_in, sort_array, axis=1)

    for bi in bound_id:
        bcs[str(bi)] = np.argwhere(bcs_sorted == bi)

    dic_to_nutils.update(
        {
            "nodes": elements_sorted,
            "cnodes": elements_sorted,
            "coords": vertices,
            "tags": {},
            "btags": bcs,
            "ptags": {},
        }
    )

    return dic_to_nutils



"""gustaf/gustaf/io/__init__.py.

io.
I - `load`.
O - `export`.
"""

from gustaf.io import ioutils, meshio, mfem, mixd, nutils
from gustaf.io.default import load

__all__ = [
    "ioutils",
    "mfem",
    "meshio",
    "mixd",
    "nutils",
    "load",
]



import numpy as np

from gustaf.spline import base
from gustaf.spline.microstructure.tiles.tilebase import TileBase


class InverseCrossTile3D(TileBase):
    """Class that provides necessary functions to create inverse microtile,
    that can be used to describe the domain within a microstructure."""

    def __init__(self):
        """Simple inverse crosstile to tile with linear-quadratic branches and
        a trilinear center spline."""
        self._dim = 3
        self._evaluation_points = np.array(
            [
                [0.0, 0.5, 0.5],
                [1.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 1.0, 0.5],
                [0.5, 0.5, 0.0],
                [0.5, 0.5, 1.0],
            ]
        )
        self._parameter_space_dimension = 1

    def closing_tile(
        self,
        parameters=None,
        closure=None,
        boundary_width=0.1,
        filling_height=0.5,
        seperator_distance=None,
        **kwargs,
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters : tuple(np.ndarray)
          radii of fitting cylinder at evaluation points
        closure : str
          parametric dimension that needs to be closed.
          Must be one of {"z_min", "z_max"}
        boundary_width : float
          with of the boundary surronding branch
        filling_height : float
          portion of the height that is filled in parametric domain

        Returns
        -------
        list_of_splines : list
        """
        # Check parameters
        if closure is None:
            raise ValueError("No closing direction given")

        if seperator_distance is None:
            raise ValueError(
                "Seperator Distance is missing. The value is required to "
                "create watertight connections with neighboring elements."
                " The value should be greater than the biggest branch "
                "radius"
            )

        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        parameters = parameters[0]
        if not (np.all(parameters > 0) and np.all(parameters < 0.5)):
            raise ValueError("Thickness out of range (0, .5)")

        if not (0.0 < float(boundary_width) < 0.5):
            raise ValueError("Boundary Width is out of range")

        if not (0.0 < float(filling_height) < 1.0):
            raise ValueError("Filling must  be in (0,1)")

        # Precompute auxiliary values
        inv_filling_height = 1.0 - filling_height
        ctps_mid_height_top = (1 + filling_height) * 0.5
        ctps_mid_height_bottom = 1.0 - ctps_mid_height_top
        center_width = 1.0 - 2 * boundary_width
        r_center = center_width * 0.5
        half_r_center = (r_center + 0.5) * 0.5
        aux_column_width = 0.5 - 2 * (0.5 - seperator_distance)

        spline_list = []
        if closure == "z_min":
            branch_thickness = parameters[5]
            branch_neighbor_x_min_ctps = np.array(
                [
                    [-0.5, -r_center, filling_height],
                    [-half_r_center, -r_center, filling_height],
                    [-r_center, -r_center, filling_height],
                    [-0.5, r_center, filling_height],
                    [-half_r_center, r_center, filling_height],
                    [-r_center, r_center, filling_height],
                    [-0.5, -aux_column_width, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_top,
                    ],
                    [-0.5, aux_column_width, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [-branch_thickness, branch_thickness, ctps_mid_height_top],
                    [-0.5, -aux_column_width, 1.0],
                    [-seperator_distance, -aux_column_width, 1.0],
                    [-branch_thickness, -branch_thickness, 1.0],
                    [-0.5, aux_column_width, 1.0],
                    [-seperator_distance, aux_column_width, 1.0],
                    [-branch_thickness, branch_thickness, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 1, 2],
                    control_points=branch_neighbor_x_min_ctps,
                )
            )

            branch_neighbor_x_max_ctps = np.array(
                [
                    [r_center, -r_center, filling_height],
                    [half_r_center, -r_center, filling_height],
                    [0.5, -r_center, filling_height],
                    [r_center, r_center, filling_height],
                    [half_r_center, r_center, filling_height],
                    [0.5, r_center, filling_height],
                    [branch_thickness, -branch_thickness, ctps_mid_height_top],
                    [
                        seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [0.5, -aux_column_width, ctps_mid_height_top],
                    [branch_thickness, branch_thickness, ctps_mid_height_top],
                    [
                        seperator_distance,
                        aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [0.5, aux_column_width, ctps_mid_height_top],
                    [branch_thickness, -branch_thickness, 1.0],
                    [seperator_distance, -aux_column_width, 1.0],
                    [0.5, -aux_column_width, 1.0],
                    [branch_thickness, branch_thickness, 1.0],
                    [seperator_distance, aux_column_width, 1.0],
                    [0.5, aux_column_width, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 1, 2],
                    control_points=branch_neighbor_x_max_ctps,
                )
            )

            branch_neighbor_y_min_ctps = np.array(
                [
                    [-r_center, -0.5, filling_height],
                    [r_center, -0.5, filling_height],
                    [-r_center, -half_r_center, filling_height],
                    [r_center, -half_r_center, filling_height],
                    [-r_center, -r_center, filling_height],
                    [r_center, -r_center, filling_height],
                    [-aux_column_width, -0.5, ctps_mid_height_top],
                    [aux_column_width, -0.5, ctps_mid_height_top],
                    [
                        -aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_top,
                    ],
                    [branch_thickness, -branch_thickness, ctps_mid_height_top],
                    [-aux_column_width, -0.5, 1.0],
                    [aux_column_width, -0.5, 1.0],
                    [-aux_column_width, -seperator_distance, 1.0],
                    [aux_column_width, -seperator_distance, 1.0],
                    [-branch_thickness, -branch_thickness, 1.0],
                    [branch_thickness, -branch_thickness, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[1, 2, 2],
                    control_points=branch_neighbor_y_min_ctps,
                )
            )

            branch_neighbor_y_max_ctps = np.array(
                [
                    [-r_center, r_center, filling_height],
                    [r_center, r_center, filling_height],
                    [-r_center, half_r_center, filling_height],
                    [r_center, half_r_center, filling_height],
                    [-r_center, 0.5, filling_height],
                    [r_center, 0.5, filling_height],
                    [-branch_thickness, branch_thickness, ctps_mid_height_top],
                    [branch_thickness, branch_thickness, ctps_mid_height_top],
                    [
                        -aux_column_width,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        aux_column_width,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [-aux_column_width, 0.5, ctps_mid_height_top],
                    [aux_column_width, 0.5, ctps_mid_height_top],
                    [-branch_thickness, branch_thickness, 1.0],
                    [branch_thickness, branch_thickness, 1.0],
                    [-aux_column_width, seperator_distance, 1.0],
                    [aux_column_width, seperator_distance, 1.0],
                    [-aux_column_width, 0.5, 1.0],
                    [aux_column_width, 0.5, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[1, 2, 2],
                    control_points=branch_neighbor_y_max_ctps,
                )
            )

            branch_x_min_y_min_ctps = np.array(
                [
                    [-0.5, -0.5, filling_height],
                    [-half_r_center, -0.5, filling_height],
                    [-r_center, -0.5, filling_height],
                    [-0.5, -half_r_center, filling_height],
                    [-half_r_center, -half_r_center, filling_height],
                    [-r_center, -half_r_center, filling_height],
                    [-0.5, -r_center, filling_height],
                    [-half_r_center, -r_center, filling_height],
                    [-r_center, -r_center, filling_height],
                    [-0.5, -0.5, ctps_mid_height_top],
                    [-seperator_distance, -0.5, ctps_mid_height_top],
                    [-aux_column_width, -0.5, ctps_mid_height_top],
                    [-0.5, -seperator_distance, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        -aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [-0.5, -aux_column_width, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_top,
                    ],
                    [-0.5, -0.5, 1.0],
                    [-seperator_distance, -0.5, 1.0],
                    [-aux_column_width, -0.5, 1.0],
                    [-0.5, -seperator_distance, 1.0],
                    [-seperator_distance, -seperator_distance, 1.0],
                    [-aux_column_width, -seperator_distance, 1.0],
                    [-0.5, -aux_column_width, 1.0],
                    [-seperator_distance, -aux_column_width, 1.0],
                    [-branch_thickness, -branch_thickness, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_min_y_min_ctps
                )
            )

            branch_x_min_y_max_ctps = np.array(
                [
                    [-0.5, r_center, filling_height],
                    [-half_r_center, r_center, filling_height],
                    [-r_center, r_center, filling_height],
                    [-0.5, half_r_center, filling_height],
                    [-half_r_center, half_r_center, filling_height],
                    [-r_center, half_r_center, filling_height],
                    [-0.5, 0.5, filling_height],
                    [-half_r_center, 0.5, filling_height],
                    [-r_center, 0.5, filling_height],
                    [-0.5, aux_column_width, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [-branch_thickness, branch_thickness, ctps_mid_height_top],
                    [-0.5, seperator_distance, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        -aux_column_width,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [-0.5, 0.5, ctps_mid_height_top],
                    [-seperator_distance, 0.5, ctps_mid_height_top],
                    [-aux_column_width, 0.5, ctps_mid_height_top],
                    [-0.5, aux_column_width, 1.0],
                    [-seperator_distance, aux_column_width, 1.0],
                    [-branch_thickness, branch_thickness, 1.0],
                    [-0.5, seperator_distance, 1.0],
                    [-seperator_distance, seperator_distance, 1.0],
                    [-aux_column_width, seperator_distance, 1.0],
                    [-0.5, 0.5, 1.0],
                    [-seperator_distance, 0.5, 1.0],
                    [-aux_column_width, 0.5, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_min_y_max_ctps
                )
            )

            branch_x_max_y_min_ctps = np.array(
                [
                    [r_center, -0.5, filling_height],
                    [half_r_center, -0.5, filling_height],
                    [0.5, -0.5, filling_height],
                    [r_center, -half_r_center, filling_height],
                    [half_r_center, -half_r_center, filling_height],
                    [0.5, -half_r_center, filling_height],
                    [r_center, -r_center, filling_height],
                    [half_r_center, -r_center, filling_height],
                    [0.5, -r_center, filling_height],
                    [aux_column_width, -0.5, ctps_mid_height_top],
                    [seperator_distance, -0.5, ctps_mid_height_top],
                    [0.5, -0.5, ctps_mid_height_top],
                    [
                        aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        seperator_distance,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [0.5, -seperator_distance, ctps_mid_height_top],
                    [branch_thickness, -branch_thickness, ctps_mid_height_top],
                    [
                        seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [0.5, -aux_column_width, ctps_mid_height_top],
                    [aux_column_width, -0.5, 1.0],
                    [seperator_distance, -0.5, 1.0],
                    [0.5, -0.5, 1.0],
                    [aux_column_width, -seperator_distance, 1.0],
                    [seperator_distance, -seperator_distance, 1.0],
                    [0.5, -seperator_distance, 1.0],
                    [branch_thickness, -branch_thickness, 1.0],
                    [seperator_distance, -aux_column_width, 1.0],
                    [0.5, -aux_column_width, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_max_y_min_ctps
                )
            )

            branch_x_max_y_max_ctps = np.array(
                [
                    [r_center, r_center, filling_height],
                    [half_r_center, r_center, filling_height],
                    [0.5, r_center, filling_height],
                    [r_center, half_r_center, filling_height],
                    [half_r_center, half_r_center, filling_height],
                    [0.5, half_r_center, filling_height],
                    [r_center, 0.5, filling_height],
                    [half_r_center, 0.5, filling_height],
                    [0.5, 0.5, filling_height],
                    [branch_thickness, branch_thickness, ctps_mid_height_top],
                    [
                        seperator_distance,
                        aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [0.5, aux_column_width, ctps_mid_height_top],
                    [
                        aux_column_width,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        seperator_distance,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [0.5, seperator_distance, ctps_mid_height_top],
                    [aux_column_width, 0.5, ctps_mid_height_top],
                    [seperator_distance, 0.5, ctps_mid_height_top],
                    [0.5, 0.5, ctps_mid_height_top],
                    [branch_thickness, branch_thickness, 1.0],
                    [seperator_distance, aux_column_width, 1.0],
                    [0.5, aux_column_width, 1.0],
                    [aux_column_width, seperator_distance, 1.0],
                    [seperator_distance, seperator_distance, 1.0],
                    [0.5, seperator_distance, 1.0],
                    [aux_column_width, 0.5, 1.0],
                    [seperator_distance, 0.5, 1.0],
                    [0.5, 0.5, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_max_y_max_ctps
                )
            )

            return spline_list

        elif closure == "z_max":
            branch_thickness = parameters[4]
            branch_neighbor_x_min_ctps = np.array(
                [
                    [-0.5, -aux_column_width, 0.0],
                    [-seperator_distance, -aux_column_width, 0.0],
                    [-branch_thickness, -branch_thickness, 0.0],
                    [-0.5, aux_column_width, 0.0],
                    [-seperator_distance, aux_column_width, 0.0],
                    [-branch_thickness, branch_thickness, 0.0],
                    [-0.5, -aux_column_width, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, aux_column_width, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, -r_center, inv_filling_height],
                    [-half_r_center, -r_center, inv_filling_height],
                    [-r_center, -r_center, inv_filling_height],
                    [-0.5, r_center, inv_filling_height],
                    [-half_r_center, r_center, inv_filling_height],
                    [-r_center, r_center, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 1, 2],
                    control_points=branch_neighbor_x_min_ctps,
                )
            )

            branch_neighbor_x_max_ctps = np.array(
                [
                    [branch_thickness, -branch_thickness, 0.0],
                    [seperator_distance, -aux_column_width, 0.0],
                    [0.5, -aux_column_width, 0.0],
                    [branch_thickness, branch_thickness, 0.0],
                    [seperator_distance, aux_column_width, 0.0],
                    [0.5, aux_column_width, 0.0],
                    [
                        branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, -aux_column_width, ctps_mid_height_bottom],
                    [
                        branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, aux_column_width, ctps_mid_height_bottom],
                    [r_center, -r_center, inv_filling_height],
                    [half_r_center, -r_center, inv_filling_height],
                    [0.5, -r_center, inv_filling_height],
                    [r_center, r_center, inv_filling_height],
                    [half_r_center, r_center, inv_filling_height],
                    [0.5, r_center, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 1, 2],
                    control_points=branch_neighbor_x_max_ctps,
                )
            )

            branch_neighbor_y_min_ctps = np.array(
                [
                    [-aux_column_width, -0.5, 0.0],
                    [aux_column_width, -0.5, 0.0],
                    [-aux_column_width, -seperator_distance, 0.0],
                    [aux_column_width, -seperator_distance, 0.0],
                    [-branch_thickness, -branch_thickness, 0.0],
                    [branch_thickness, -branch_thickness, 0.0],
                    [-aux_column_width, -0.5, ctps_mid_height_bottom],
                    [aux_column_width, -0.5, ctps_mid_height_bottom],
                    [
                        -aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-r_center, -0.5, inv_filling_height],
                    [r_center, -0.5, inv_filling_height],
                    [-r_center, -half_r_center, inv_filling_height],
                    [r_center, -half_r_center, inv_filling_height],
                    [-r_center, -r_center, inv_filling_height],
                    [r_center, -r_center, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[1, 2, 2],
                    control_points=branch_neighbor_y_min_ctps,
                )
            )

            branch_neighbor_y_max_ctps = np.array(
                [
                    [-branch_thickness, branch_thickness, 0.0],
                    [branch_thickness, branch_thickness, 0.0],
                    [-aux_column_width, seperator_distance, 0.0],
                    [aux_column_width, seperator_distance, 0.0],
                    [-aux_column_width, 0.5, 0.0],
                    [aux_column_width, 0.5, 0.0],
                    [
                        -branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -aux_column_width,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        aux_column_width,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [-aux_column_width, 0.5, ctps_mid_height_bottom],
                    [aux_column_width, 0.5, ctps_mid_height_bottom],
                    [-r_center, r_center, inv_filling_height],
                    [r_center, r_center, inv_filling_height],
                    [-r_center, half_r_center, inv_filling_height],
                    [r_center, half_r_center, inv_filling_height],
                    [-r_center, 0.5, inv_filling_height],
                    [r_center, 0.5, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[1, 2, 2],
                    control_points=branch_neighbor_y_max_ctps,
                )
            )

            branch_x_min_y_min_ctps = np.array(
                [
                    [-0.5, -0.5, 0.0],
                    [-seperator_distance, -0.5, 0.0],
                    [-aux_column_width, -0.5, 0.0],
                    [-0.5, -seperator_distance, 0.0],
                    [-seperator_distance, -seperator_distance, 0.0],
                    [-aux_column_width, -seperator_distance, 0.0],
                    [-0.5, -aux_column_width, 0.0],
                    [-seperator_distance, -aux_column_width, 0.0],
                    [-branch_thickness, -branch_thickness, 0.0],
                    [-0.5, -0.5, ctps_mid_height_bottom],
                    [-seperator_distance, -0.5, ctps_mid_height_bottom],
                    [-aux_column_width, -0.5, ctps_mid_height_bottom],
                    [-0.5, -seperator_distance, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, -aux_column_width, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, -0.5, inv_filling_height],
                    [-half_r_center, -0.5, inv_filling_height],
                    [-r_center, -0.5, inv_filling_height],
                    [-0.5, -half_r_center, inv_filling_height],
                    [-half_r_center, -half_r_center, inv_filling_height],
                    [-r_center, -half_r_center, inv_filling_height],
                    [-0.5, -r_center, inv_filling_height],
                    [-half_r_center, -r_center, inv_filling_height],
                    [-r_center, -r_center, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_min_y_min_ctps
                )
            )

            branch_x_max_y_max_ctps = np.array(
                [
                    [branch_thickness, branch_thickness, 0.0],
                    [seperator_distance, aux_column_width, 0.0],
                    [0.5, aux_column_width, 0.0],
                    [aux_column_width, seperator_distance, 0.0],
                    [seperator_distance, seperator_distance, 0.0],
                    [0.5, seperator_distance, 0.0],
                    [aux_column_width, 0.5, 0.0],
                    [seperator_distance, 0.5, 0.0],
                    [0.5, 0.5, 0.0],
                    [
                        branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, aux_column_width, ctps_mid_height_bottom],
                    [
                        aux_column_width,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, seperator_distance, ctps_mid_height_bottom],
                    [aux_column_width, 0.5, ctps_mid_height_bottom],
                    [seperator_distance, 0.5, ctps_mid_height_bottom],
                    [0.5, 0.5, ctps_mid_height_bottom],
                    [r_center, r_center, inv_filling_height],
                    [half_r_center, r_center, inv_filling_height],
                    [0.5, r_center, inv_filling_height],
                    [r_center, half_r_center, inv_filling_height],
                    [half_r_center, half_r_center, inv_filling_height],
                    [0.5, half_r_center, inv_filling_height],
                    [r_center, 0.5, inv_filling_height],
                    [half_r_center, 0.5, inv_filling_height],
                    [0.5, 0.5, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_max_y_max_ctps
                )
            )

            branch_x_max_y_min_ctps = np.array(
                [
                    [aux_column_width, -0.5, 0.0],
                    [seperator_distance, -0.5, 0.0],
                    [0.5, -0.5, 0.0],
                    [aux_column_width, -seperator_distance, 0.0],
                    [seperator_distance, -seperator_distance, 0.0],
                    [0.5, -seperator_distance, 0.0],
                    [branch_thickness, -branch_thickness, 0.0],
                    [seperator_distance, -aux_column_width, 0.0],
                    [0.5, -aux_column_width, 0.0],
                    [aux_column_width, -0.5, ctps_mid_height_bottom],
                    [seperator_distance, -0.5, ctps_mid_height_bottom],
                    [0.5, -0.5, ctps_mid_height_bottom],
                    [
                        aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, -seperator_distance, ctps_mid_height_bottom],
                    [
                        branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, -aux_column_width, ctps_mid_height_bottom],
                    [r_center, -0.5, inv_filling_height],
                    [half_r_center, -0.5, inv_filling_height],
                    [0.5, -0.5, inv_filling_height],
                    [r_center, -half_r_center, inv_filling_height],
                    [half_r_center, -half_r_center, inv_filling_height],
                    [0.5, -half_r_center, inv_filling_height],
                    [r_center, -r_center, inv_filling_height],
                    [half_r_center, -r_center, inv_filling_height],
                    [0.5, -r_center, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_max_y_min_ctps
                )
            )

            branch_x_min_y_max_ctps = np.array(
                [
                    [-0.5, aux_column_width, 0.0],
                    [-seperator_distance, aux_column_width, 0.0],
                    [-branch_thickness, branch_thickness, 0.0],
                    [-0.5, seperator_distance, 0.0],
                    [-seperator_distance, seperator_distance, 0.0],
                    [-aux_column_width, seperator_distance, 0.0],
                    [-0.5, 0.5, 0.0],
                    [-seperator_distance, 0.5, 0.0],
                    [-aux_column_width, 0.5, 0.0],
                    [-0.5, aux_column_width, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, seperator_distance, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -aux_column_width,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, 0.5, ctps_mid_height_bottom],
                    [-seperator_distance, 0.5, ctps_mid_height_bottom],
                    [-aux_column_width, 0.5, ctps_mid_height_bottom],
                    [-0.5, r_center, inv_filling_height],
                    [-half_r_center, r_center, inv_filling_height],
                    [-r_center, r_center, inv_filling_height],
                    [-0.5, half_r_center, inv_filling_height],
                    [-half_r_center, half_r_center, inv_filling_height],
                    [-r_center, half_r_center, inv_filling_height],
                    [-0.5, 0.5, inv_filling_height],
                    [-half_r_center, 0.5, inv_filling_height],
                    [-r_center, 0.5, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_min_y_max_ctps
                )
            )

            return spline_list
        else:
            raise ValueError("Corner Type not supported")

    def create_tile(
        self,
        parameters=None,
        seperator_distance=None,
        center_expansion=1.0,
        **kwargs,
    ):
        """Create an inverse microtile based on the parameters that describe
        the branch thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : tuple(np.array)
          only first entry is used, defines the internal radii of the
          branches
        seperator_distance : float
          position of the control points for higher order elements
        center_expansion : float
          thickness of center is expanded by a factor

        Returns
        -------
        microtile_list : list(splines)
        """

        if not isinstance(center_expansion, float):
            raise ValueError("Invalid Type")
        if not ((center_expansion > 0.5) and (center_expansion < 1.5)):
            raise ValueError("Center Expansion must be in (.5, 1.5)")

        if seperator_distance is None:
            raise ValueError(
                "Seperator Distance is missing. The value is required to "
                "create watertight connections with neighboring elements."
                " The value should be greater than the biggest branch "
                "radius"
            )

        # Check if all radii are in allowed range
        max_radius = min(0.5, (0.5 / center_expansion))
        max_radius = min(max_radius, seperator_distance)

        # set to default if nothing is given
        if parameters is None:
            self._logd("Setting branch thickness to default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r, z_max_r] = parameters[
            0
        ].tolist()

        for radius in [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r, z_max_r]:
            if not isinstance(radius, float):
                raise ValueError("Invalid type")
            if not (radius > 0 and radius < max_radius):
                raise ValueError(
                    f"Radii must be in (0,{max_radius}) for "
                    f"center_expansion {center_expansion}"
                )

        # center radius
        center_r = (
            (x_min_r + x_max_r + y_min_r + y_max_r + z_min_r + z_max_r)
            / 6.0
            * center_expansion
        )

        # Auxiliary values for smooothing (mid-branch thickness)
        aux_x_min = min(x_min_r, center_r)
        aux_x_max = min(x_max_r, center_r)
        aux_y_min = min(y_min_r, center_r)
        aux_y_max = min(y_max_r, center_r)
        aux_z_min = min(z_min_r, center_r)
        aux_z_max = min(z_max_r, center_r)
        # Branch midlength
        hd_center = 0.5 * (0.5 + center_r)

        #
        if seperator_distance is None:
            seperator_distance = 0.45
        aux_column_width = 0.5 - 2 * (0.5 - seperator_distance)

        # Init return type
        spline_list = []

        # Start with branch interconnections
        x_min_y_min = np.array(
            [
                [-0.5, -0.5, -aux_column_width],
                [-seperator_distance, -0.5, -aux_column_width],
                [-y_min_r, -0.5, -y_min_r],
                [-0.5, -seperator_distance, -aux_column_width],
                [-hd_center, -hd_center, -aux_column_width],
                [-aux_y_min, -hd_center, -aux_y_min],
                [-0.5, -x_min_r, -x_min_r],
                [-hd_center, -aux_x_min, -aux_x_min],
                [-center_r, -center_r, -center_r],
                [-0.5, -0.5, aux_column_width],
                [-seperator_distance, -0.5, aux_column_width],
                [-y_min_r, -0.5, y_min_r],
                [-0.5, -seperator_distance, aux_column_width],
                [-hd_center, -hd_center, aux_column_width],
                [-aux_y_min, -hd_center, aux_y_min],
                [-0.5, -x_min_r, x_min_r],
                [-hd_center, -aux_x_min, aux_x_min],
                [-center_r, -center_r, center_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 1], control_points=x_min_y_min)
        )

        x_max_y_min = np.array(
            [
                [y_min_r, -0.5, -y_min_r],
                [seperator_distance, -0.5, -aux_column_width],
                [0.5, -0.5, -aux_column_width],
                [aux_y_min, -hd_center, -aux_y_min],
                [hd_center, -hd_center, -aux_column_width],
                [0.5, -seperator_distance, -aux_column_width],
                [center_r, -center_r, -center_r],
                [hd_center, -aux_x_max, -aux_x_max],
                [0.5, -x_max_r, -x_max_r],
                [y_min_r, -0.5, y_min_r],
                [seperator_distance, -0.5, aux_column_width],
                [0.5, -0.5, aux_column_width],
                [aux_y_min, -hd_center, aux_y_min],
                [hd_center, -hd_center, aux_column_width],
                [0.5, -seperator_distance, aux_column_width],
                [center_r, -center_r, center_r],
                [hd_center, -aux_x_max, aux_x_max],
                [0.5, -x_max_r, x_max_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 1], control_points=x_max_y_min)
        )

        x_min_y_max = np.array(
            [
                [-0.5, x_min_r, -x_min_r],
                [-hd_center, aux_x_min, -aux_x_min],
                [-center_r, center_r, -center_r],
                [-0.5, seperator_distance, -aux_column_width],
                [-hd_center, hd_center, -aux_column_width],
                [-aux_y_max, hd_center, -aux_y_max],
                [-0.5, 0.5, -aux_column_width],
                [-seperator_distance, 0.5, -aux_column_width],
                [-y_max_r, 0.5, -y_max_r],
                [-0.5, x_min_r, x_min_r],
                [-hd_center, aux_x_min, aux_x_min],
                [-center_r, center_r, center_r],
                [-0.5, seperator_distance, aux_column_width],
                [-hd_center, hd_center, aux_column_width],
                [-aux_y_max, hd_center, aux_y_max],
                [-0.5, 0.5, aux_column_width],
                [-seperator_distance, 0.5, aux_column_width],
                [-y_max_r, 0.5, y_max_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 1], control_points=x_min_y_max)
        )

        x_max_y_max = np.array(
            [
                [center_r, center_r, -center_r],
                [hd_center, aux_x_max, -aux_x_max],
                [0.5, x_max_r, -x_max_r],
                [aux_y_max, hd_center, -aux_y_max],
                [hd_center, hd_center, -aux_column_width],
                [0.5, seperator_distance, -aux_column_width],
                [y_max_r, 0.5, -y_max_r],
                [seperator_distance, 0.5, -aux_column_width],
                [0.5, 0.5, -aux_column_width],
                [center_r, center_r, center_r],
                [hd_center, aux_x_max, aux_x_max],
                [0.5, x_max_r, x_max_r],
                [aux_y_max, hd_center, aux_y_max],
                [hd_center, hd_center, aux_column_width],
                [0.5, seperator_distance, aux_column_width],
                [y_max_r, 0.5, y_max_r],
                [seperator_distance, 0.5, aux_column_width],
                [0.5, 0.5, aux_column_width],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 1], control_points=x_max_y_max)
        )

        x_min_z_min = np.array(
            [
                [-0.5, -aux_column_width, -0.5],
                [-seperator_distance, -aux_column_width, -0.5],
                [-z_min_r, -z_min_r, -0.5],
                [-0.5, aux_column_width, -0.5],
                [-seperator_distance, aux_column_width, -0.5],
                [-z_min_r, z_min_r, -0.5],
                [-0.5, -aux_column_width, -seperator_distance],
                [-hd_center, -aux_column_width, -hd_center],
                [-aux_z_min, -aux_z_min, -hd_center],
                [-0.5, aux_column_width, -seperator_distance],
                [-hd_center, aux_column_width, -hd_center],
                [-aux_z_min, aux_z_min, -hd_center],
                [-0.5, -x_min_r, -x_min_r],
                [-hd_center, -aux_x_min, -aux_x_min],
                [-center_r, -center_r, -center_r],
                [-0.5, x_min_r, -x_min_r],
                [-hd_center, aux_x_min, -aux_x_min],
                [-center_r, center_r, -center_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 1, 2], control_points=x_min_z_min)
        )

        x_max_z_min = np.array(
            [
                [z_min_r, -z_min_r, -0.5],
                [seperator_distance, -aux_column_width, -0.5],
                [0.5, -aux_column_width, -0.5],
                [z_min_r, z_min_r, -0.5],
                [seperator_distance, aux_column_width, -0.5],
                [0.5, aux_column_width, -0.5],
                [aux_z_min, -aux_z_min, -hd_center],
                [hd_center, -aux_column_width, -hd_center],
                [0.5, -aux_column_width, -seperator_distance],
                [aux_z_min, aux_z_min, -hd_center],
                [hd_center, aux_column_width, -hd_center],
                [0.5, aux_column_width, -seperator_distance],
                [center_r, -center_r, -center_r],
                [hd_center, -aux_x_max, -aux_x_max],
                [0.5, -x_max_r, -x_max_r],
                [center_r, center_r, -center_r],
                [hd_center, aux_x_max, -aux_x_max],
                [0.5, x_max_r, -x_max_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 1, 2], control_points=x_max_z_min)
        )

        x_min_z_max = np.array(
            [
                [-0.5, -x_min_r, x_min_r],
                [-hd_center, -aux_x_min, aux_x_min],
                [-center_r, -center_r, center_r],
                [-0.5, x_min_r, x_min_r],
                [-hd_center, aux_x_min, aux_x_min],
                [-center_r, center_r, center_r],
                [-0.5, -aux_column_width, seperator_distance],
                [-hd_center, -aux_column_width, hd_center],
                [-aux_z_max, -aux_z_max, hd_center],
                [-0.5, aux_column_width, seperator_distance],
                [-hd_center, aux_column_width, hd_center],
                [-aux_z_max, aux_z_max, hd_center],
                [-0.5, -aux_column_width, 0.5],
                [-seperator_distance, -aux_column_width, 0.5],
                [-z_max_r, -z_max_r, 0.5],
                [-0.5, aux_column_width, 0.5],
                [-seperator_distance, aux_column_width, 0.5],
                [-z_max_r, z_max_r, 0.5],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 1, 2], control_points=x_min_z_max)
        )

        x_max_z_max = np.array(
            [
                [center_r, -center_r, center_r],
                [hd_center, -aux_x_max, aux_x_max],
                [0.5, -x_max_r, x_max_r],
                [center_r, center_r, center_r],
                [hd_center, aux_x_max, aux_x_max],
                [0.5, x_max_r, x_max_r],
                [aux_z_max, -aux_z_max, hd_center],
                [hd_center, -aux_column_width, hd_center],
                [0.5, -aux_column_width, seperator_distance],
                [aux_z_max, aux_z_max, hd_center],
                [hd_center, aux_column_width, hd_center],
                [0.5, aux_column_width, seperator_distance],
                [z_max_r, -z_max_r, 0.5],
                [seperator_distance, -aux_column_width, 0.5],
                [0.5, -aux_column_width, 0.5],
                [z_max_r, z_max_r, 0.5],
                [seperator_distance, aux_column_width, 0.5],
                [0.5, aux_column_width, 0.5],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 1, 2], control_points=x_max_z_max)
        )

        y_min_z_min = np.array(
            [
                [-aux_column_width, -0.5, -0.5],
                [aux_column_width, -0.5, -0.5],
                [-aux_column_width, -seperator_distance, -0.5],
                [aux_column_width, -seperator_distance, -0.5],
                [-z_min_r, -z_min_r, -0.5],
                [z_min_r, -z_min_r, -0.5],
                [-aux_column_width, -0.5, -seperator_distance],
                [aux_column_width, -0.5, -seperator_distance],
                [-aux_column_width, -hd_center, -hd_center],
                [aux_column_width, -hd_center, -hd_center],
                [-aux_z_min, -aux_z_min, -hd_center],
                [aux_z_min, -aux_z_min, -hd_center],
                [-y_min_r, -0.5, -y_min_r],
                [y_min_r, -0.5, -y_min_r],
                [-aux_y_min, -hd_center, -aux_y_min],
                [aux_y_min, -hd_center, -aux_y_min],
                [-center_r, -center_r, -center_r],
                [center_r, -center_r, -center_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[1, 2, 2], control_points=y_min_z_min)
        )

        y_max_z_min = np.array(
            [
                [-z_min_r, z_min_r, -0.5],
                [z_min_r, z_min_r, -0.5],
                [-aux_column_width, seperator_distance, -0.5],
                [aux_column_width, seperator_distance, -0.5],
                [-aux_column_width, 0.5, -0.5],
                [aux_column_width, 0.5, -0.5],
                [-aux_z_min, aux_z_min, -hd_center],
                [aux_z_min, aux_z_min, -hd_center],
                [-aux_column_width, hd_center, -hd_center],
                [aux_column_width, hd_center, -hd_center],
                [-aux_column_width, 0.5, -seperator_distance],
                [aux_column_width, 0.5, -seperator_distance],
                [-center_r, center_r, -center_r],
                [center_r, center_r, -center_r],
                [-aux_y_max, hd_center, -aux_y_max],
                [aux_y_max, hd_center, -aux_y_max],
                [-y_max_r, 0.5, -y_max_r],
                [y_max_r, 0.5, -y_max_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[1, 2, 2], control_points=y_max_z_min)
        )

        y_min_z_max = np.array(
            [
                [-y_min_r, -0.5, y_min_r],
                [y_min_r, -0.5, y_min_r],
                [-aux_y_min, -hd_center, aux_y_min],
                [aux_y_min, -hd_center, aux_y_min],
                [-center_r, -center_r, center_r],
                [center_r, -center_r, center_r],
                [-aux_column_width, -0.5, seperator_distance],
                [aux_column_width, -0.5, seperator_distance],
                [-aux_column_width, -hd_center, hd_center],
                [aux_column_width, -hd_center, hd_center],
                [-aux_z_max, -aux_z_max, hd_center],
                [aux_z_max, -aux_z_max, hd_center],
                [-aux_column_width, -0.5, 0.5],
                [aux_column_width, -0.5, 0.5],
                [-aux_column_width, -seperator_distance, 0.5],
                [aux_column_width, -seperator_distance, 0.5],
                [-z_max_r, -z_max_r, 0.5],
                [z_max_r, -z_max_r, 0.5],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[1, 2, 2], control_points=y_min_z_max)
        )

        y_max_z_max = np.array(
            [
                [-center_r, center_r, center_r],
                [center_r, center_r, center_r],
                [-aux_y_max, hd_center, aux_y_max],
                [aux_y_max, hd_center, aux_y_max],
                [-y_max_r, 0.5, y_max_r],
                [y_max_r, 0.5, y_max_r],
                [-aux_z_max, aux_z_max, hd_center],
                [aux_z_max, aux_z_max, hd_center],
                [-aux_column_width, hd_center, hd_center],
                [aux_column_width, hd_center, hd_center],
                [-aux_column_width, 0.5, seperator_distance],
                [aux_column_width, 0.5, seperator_distance],
                [-z_max_r, z_max_r, 0.5],
                [z_max_r, z_max_r, 0.5],
                [-aux_column_width, seperator_distance, 0.5],
                [aux_column_width, seperator_distance, 0.5],
                [-aux_column_width, 0.5, 0.5],
                [aux_column_width, 0.5, 0.5],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[1, 2, 2], control_points=y_max_z_max)
        )

        x_min_y_min_z_min = np.array(
            [
                [-0.5, -0.5, -0.5],
                [-seperator_distance, -0.5, -0.5],
                [-aux_column_width, -0.5, -0.5],
                [-0.5, -seperator_distance, -0.5],
                [-seperator_distance, -seperator_distance, -0.5],
                [-aux_column_width, -seperator_distance, -0.5],
                [-0.5, -aux_column_width, -0.5],
                [-seperator_distance, -aux_column_width, -0.5],
                [-z_min_r, -z_min_r, -0.5],
                [-0.5, -0.5, -seperator_distance],
                [-seperator_distance, -0.5, -seperator_distance],
                [-aux_column_width, -0.5, -seperator_distance],
                [-0.5, -seperator_distance, -seperator_distance],
                [-hd_center, -hd_center, -hd_center],
                [-aux_column_width, -hd_center, -hd_center],
                [-0.5, -aux_column_width, -seperator_distance],
                [-hd_center, -aux_column_width, -hd_center],
                [-aux_z_min, -aux_z_min, -hd_center],
                [-0.5, -0.5, -aux_column_width],
                [-seperator_distance, -0.5, -aux_column_width],
                [-y_min_r, -0.5, -y_min_r],
                [-0.5, -seperator_distance, -aux_column_width],
                [-hd_center, -hd_center, -aux_column_width],
                [-aux_y_min, -hd_center, -aux_y_min],
                [-0.5, -x_min_r, -x_min_r],
                [-hd_center, -aux_x_min, -aux_x_min],
                [-center_r, -center_r, -center_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 2], control_points=x_min_y_min_z_min)
        )

        x_max_y_min_z_min = np.array(
            [
                [aux_column_width, -0.5, -0.5],
                [seperator_distance, -0.5, -0.5],
                [0.5, -0.5, -0.5],
                [aux_column_width, -seperator_distance, -0.5],
                [seperator_distance, -seperator_distance, -0.5],
                [0.5, -seperator_distance, -0.5],
                [z_min_r, -z_min_r, -0.5],
                [seperator_distance, -aux_column_width, -0.5],
                [0.5, -aux_column_width, -0.5],
                [aux_column_width, -0.5, -seperator_distance],
                [seperator_distance, -0.5, -seperator_distance],
                [0.5, -0.5, -seperator_distance],
                [aux_column_width, -hd_center, -hd_center],
                [hd_center, -hd_center, -hd_center],
                [0.5, -seperator_distance, -seperator_distance],
                [aux_z_min, -aux_z_min, -hd_center],
                [hd_center, -aux_column_width, -hd_center],
                [0.5, -aux_column_width, -seperator_distance],
                [y_min_r, -0.5, -y_min_r],
                [seperator_distance, -0.5, -aux_column_width],
                [0.5, -0.5, -aux_column_width],
                [aux_y_min, -hd_center, -aux_y_min],
                [hd_center, -hd_center, -aux_column_width],
                [0.5, -seperator_distance, -aux_column_width],
                [center_r, -center_r, -center_r],
                [hd_center, -aux_x_max, -aux_x_max],
                [0.5, -x_max_r, -x_max_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 2], control_points=x_max_y_min_z_min)
        )

        x_min_y_max_z_min = np.array(
            [
                [-0.5, aux_column_width, -0.5],
                [-seperator_distance, aux_column_width, -0.5],
                [-z_min_r, z_min_r, -0.5],
                [-0.5, seperator_distance, -0.5],
                [-seperator_distance, seperator_distance, -0.5],
                [-aux_column_width, seperator_distance, -0.5],
                [-0.5, 0.5, -0.5],
                [-seperator_distance, 0.5, -0.5],
                [-aux_column_width, 0.5, -0.5],
                [-0.5, aux_column_width, -seperator_distance],
                [-hd_center, aux_column_width, -hd_center],
                [-aux_z_min, aux_z_min, -hd_center],
                [-0.5, seperator_distance, -seperator_distance],
                [-hd_center, hd_center, -hd_center],
                [-aux_column_width, hd_center, -hd_center],
                [-0.5, 0.5, -seperator_distance],
                [-seperator_distance, 0.5, -seperator_distance],
                [-aux_column_width, 0.5, -seperator_distance],
                [-0.5, x_min_r, -x_min_r],
                [-hd_center, aux_x_min, -aux_x_min],
                [-center_r, center_r, -center_r],
                [-0.5, seperator_distance, -aux_column_width],
                [-hd_center, hd_center, -aux_column_width],
                [-aux_y_max, hd_center, -aux_y_max],
                [-0.5, 0.5, -aux_column_width],
                [-seperator_distance, 0.5, -aux_column_width],
                [-y_max_r, 0.5, -y_max_r],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 2], control_points=x_min_y_max_z_min)
        )

        x_max_y_max_z_min = np.array(
            [
                [z_min_r, z_min_r, -0.5],
                [seperator_distance, aux_column_width, -0.5],
                [0.5, aux_column_width, -0.5],
                [aux_column_width, seperator_distance, -0.5],
                [seperator_distance, seperator_distance, -0.5],
                [0.5, seperator_distance, -0.5],
                [aux_column_width, 0.5, -0.5],
                [seperator_distance, 0.5, -0.5],
                [0.5, 0.5, -0.5],
                [aux_z_min, aux_z_min, -hd_center],
                [hd_center, aux_column_width, -hd_center],
                [0.5, aux_column_width, -seperator_distance],
                [aux_column_width, hd_center, -hd_center],
                [hd_center, hd_center, -hd_center],
                [0.5, seperator_distance, -seperator_distance],
                [aux_column_width, 0.5, -seperator_distance],
                [seperator_distance, 0.5, -seperator_distance],
                [0.5, 0.5, -seperator_distance],
                [center_r, center_r, -center_r],
                [hd_center, aux_x_max, -aux_x_max],
                [0.5, x_max_r, -x_max_r],
                [aux_y_max, hd_center, -aux_y_max],
                [hd_center, hd_center, -aux_column_width],
                [0.5, seperator_distance, -aux_column_width],
                [y_max_r, 0.5, -y_max_r],
                [seperator_distance, 0.5, -aux_column_width],
                [0.5, 0.5, -aux_column_width],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 2], control_points=x_max_y_max_z_min)
        )

        x_min_y_min_z_max = np.array(
            [
                [-0.5, -0.5, aux_column_width],
                [-seperator_distance, -0.5, aux_column_width],
                [-y_min_r, -0.5, y_min_r],
                [-0.5, -seperator_distance, aux_column_width],
                [-hd_center, -hd_center, aux_column_width],
                [-aux_y_min, -hd_center, aux_y_min],
                [-0.5, -x_min_r, x_min_r],
                [-hd_center, -aux_x_min, aux_x_min],
                [-center_r, -center_r, center_r],
                [-0.5, -0.5, seperator_distance],
                [-seperator_distance, -0.5, seperator_distance],
                [-aux_column_width, -0.5, seperator_distance],
                [-0.5, -seperator_distance, seperator_distance],
                [-hd_center, -hd_center, hd_center],
                [-aux_column_width, -hd_center, hd_center],
                [-0.5, -aux_column_width, seperator_distance],
                [-hd_center, -aux_column_width, hd_center],
                [-aux_z_max, -aux_z_max, hd_center],
                [-0.5, -0.5, 0.5],
                [-seperator_distance, -0.5, 0.5],
                [-aux_column_width, -0.5, 0.5],
                [-0.5, -seperator_distance, 0.5],
                [-seperator_distance, -seperator_distance, 0.5],
                [-aux_column_width, -seperator_distance, 0.5],
                [-0.5, -aux_column_width, 0.5],
                [-seperator_distance, -aux_column_width, 0.5],
                [-z_max_r, -z_max_r, 0.5],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 2], control_points=x_min_y_min_z_max)
        )

        x_max_y_min_z_max = np.array(
            [
                [y_min_r, -0.5, y_min_r],
                [seperator_distance, -0.5, aux_column_width],
                [0.5, -0.5, aux_column_width],
                [aux_y_min, -hd_center, aux_y_min],
                [hd_center, -hd_center, aux_column_width],
                [0.5, -seperator_distance, aux_column_width],
                [center_r, -center_r, center_r],
                [hd_center, -aux_x_max, aux_x_max],
                [0.5, -x_max_r, x_max_r],
                [aux_column_width, -0.5, seperator_distance],
                [seperator_distance, -0.5, seperator_distance],
                [0.5, -0.5, seperator_distance],
                [aux_column_width, -hd_center, hd_center],
                [hd_center, -hd_center, hd_center],
                [0.5, -seperator_distance, seperator_distance],
                [aux_z_max, -aux_z_max, hd_center],
                [hd_center, -aux_column_width, hd_center],
                [0.5, -aux_column_width, seperator_distance],
                [aux_column_width, -0.5, 0.5],
                [seperator_distance, -0.5, 0.5],
                [0.5, -0.5, 0.5],
                [aux_column_width, -seperator_distance, 0.5],
                [seperator_distance, -seperator_distance, 0.5],
                [0.5, -seperator_distance, 0.5],
                [z_max_r, -z_max_r, 0.5],
                [seperator_distance, -aux_column_width, 0.5],
                [0.5, -aux_column_width, 0.5],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 2], control_points=x_max_y_min_z_max)
        )

        x_min_y_max_z_max = np.array(
            [
                [-0.5, x_min_r, x_min_r],
                [-hd_center, aux_x_min, aux_x_min],
                [-center_r, center_r, center_r],
                [-0.5, seperator_distance, aux_column_width],
                [-hd_center, hd_center, aux_column_width],
                [-aux_y_max, hd_center, aux_y_max],
                [-0.5, 0.5, aux_column_width],
                [-seperator_distance, 0.5, aux_column_width],
                [-y_max_r, 0.5, y_max_r],
                [-0.5, aux_column_width, seperator_distance],
                [-hd_center, aux_column_width, hd_center],
                [-aux_z_max, aux_z_max, hd_center],
                [-0.5, seperator_distance, seperator_distance],
                [-hd_center, hd_center, hd_center],
                [-aux_column_width, hd_center, hd_center],
                [-0.5, 0.5, seperator_distance],
                [-seperator_distance, 0.5, seperator_distance],
                [-aux_column_width, 0.5, seperator_distance],
                [-0.5, aux_column_width, 0.5],
                [-seperator_distance, aux_column_width, 0.5],
                [-z_max_r, z_max_r, 0.5],
                [-0.5, seperator_distance, 0.5],
                [-seperator_distance, seperator_distance, 0.5],
                [-aux_column_width, seperator_distance, 0.5],
                [-0.5, 0.5, 0.5],
                [-seperator_distance, 0.5, 0.5],
                [-aux_column_width, 0.5, 0.5],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 2], control_points=x_min_y_max_z_max)
        )

        x_max_y_max_z_max = np.array(
            [
                [center_r, center_r, center_r],
                [hd_center, aux_x_max, aux_x_max],
                [0.5, x_max_r, x_max_r],
                [aux_y_max, hd_center, aux_y_max],
                [hd_center, hd_center, aux_column_width],
                [0.5, seperator_distance, aux_column_width],
                [y_max_r, 0.5, y_max_r],
                [seperator_distance, 0.5, aux_column_width],
                [0.5, 0.5, aux_column_width],
                [aux_z_max, aux_z_max, hd_center],
                [hd_center, aux_column_width, hd_center],
                [0.5, aux_column_width, seperator_distance],
                [aux_column_width, hd_center, hd_center],
                [hd_center, hd_center, hd_center],
                [0.5, seperator_distance, seperator_distance],
                [aux_column_width, 0.5, seperator_distance],
                [seperator_distance, 0.5, seperator_distance],
                [0.5, 0.5, seperator_distance],
                [z_max_r, z_max_r, 0.5],
                [seperator_distance, aux_column_width, 0.5],
                [0.5, aux_column_width, 0.5],
                [aux_column_width, seperator_distance, 0.5],
                [seperator_distance, seperator_distance, 0.5],
                [0.5, seperator_distance, 0.5],
                [aux_column_width, 0.5, 0.5],
                [seperator_distance, 0.5, 0.5],
                [0.5, 0.5, 0.5],
            ]
        ) + np.array([0.5, 0.5, 0.5])

        spline_list.append(
            base.Bezier(degrees=[2, 2, 2], control_points=x_max_y_max_z_max)
        )

        return spline_list



"""gustaf/gustaf/utils/connec.py.

Useful functions for connectivity operation. Ranging from edges to
volumes. Named connec because connectivity is too long. Would have been
even cooler, if it was palindrome.
"""

import numpy as np

from gustaf import helpers, settings
from gustaf.utils import arr


def tet_to_tri(volumes):
    """Computes tri faces based on following index scheme.

    ``Tetrahedron``

    .. code-block::

        Ref: (node_ind), face_ind

                      (0)
                     _/|
                   _/ 1|
            (1)  _/____| (3)
               _/|    /|
             _/ 0| 2/ 3|
            /____|/____|
          (0)   (2)    (0)

        face_ind | node_ind
        ---------|----------
        0        | 0 2 1
        1        | 1 3 0
        2        | 2 3 1
        3        | 3 2 0

    Parameters
    -----------
    volumes: (n, 4) np.ndarray

    Returns
    --------
    faces: (n * 4, 3) np.ndarray
    """
    volumes = arr.make_c_contiguous(volumes, settings.INT_DTYPE)

    if volumes.ndim != 2 or volumes.shape[1] != 4:
        raise ValueError("Given volumes are not `tet` volumes")

    fpe = 4  # faces per element
    faces = (
        np.ones(((volumes.shape[0] * fpe), 3), dtype=np.int32) * -1
    )  # -1 for safety check

    faces[:, 0] = volumes.flatten()
    faces[::fpe, [1, 2]] = volumes[:, [2, 1]]
    faces[1::fpe, [1, 2]] = volumes[:, [3, 0]]
    faces[2::fpe, [1, 2]] = volumes[:, [3, 1]]
    faces[3::fpe, [1, 2]] = volumes[:, [2, 0]]

    if (faces == -1).any():
        raise ValueError("Something went wrong while computing faces")

    return faces


def hexa_to_quad(volumes):
    """Computes quad faces based on following index scheme.

    ``Hexahedron``

    .. code-block::


                (6)    (7)
                *------*
                |      |
         (6) (2)| 3    |(3)   (7)    (6)
         *------*------*------*------*
         |      |      |      |      |
         | 2    | 0    | 4    | 5    |
         *------*------*------*------*
         (5) (1)|      |(0)   (4)    (5)
                | 1    |
                *------*
                (5)    (4)

        face_ind | node_ind
        ---------|----------
        0        | 1 0 3 2
        1        | 0 1 5 4
        2        | 1 2 6 5
        3        | 2 3 7 6
        4        | 3 0 4 7
        5        | 4 5 6 7

    Parameters
    -----------
    volumes: (n, 8) np.ndarray

    Returns
    --------
    faces: (n * 8, 4) np.ndarray
    """
    volumes = arr.make_c_contiguous(volumes, settings.INT_DTYPE)

    if volumes.ndim != 2 or volumes.shape[1] != 8:
        raise ValueError("Given volumes are not `hexa` volumes")

    fpe = 6  # faces per element
    faces = (
        np.ones(((volumes.shape[0] * fpe), 4), dtype=np.int32) * -1
    )  # -1 for safety check

    faces[::fpe] = volumes[:, [1, 0, 3, 2]]
    faces[1::fpe] = volumes[:, [0, 1, 5, 4]]
    faces[2::fpe] = volumes[:, [1, 2, 6, 5]]
    faces[3::fpe] = volumes[:, [2, 3, 7, 6]]
    faces[4::fpe] = volumes[:, [3, 0, 4, 7]]
    faces[5::fpe] = volumes[:, [4, 5, 6, 7]]

    if (faces == -1).any():
        raise ValueError("Something went wrong while computing faces")

    return faces


def volumes_to_faces(volumes):
    """Guidance function for `tet_to_tri` and `hexa_to_quad`.

    Parameters
    -----------
    volumes: (n, 4) or (m, 8) np.ndarray

    Returns
    --------
    faces: (n*4, 3) or (m*6, 4) np.ndarray
    """
    volumes = arr.make_c_contiguous(volumes, settings.INT_DTYPE)
    if volumes.shape[1] == 4:
        return tet_to_tri(volumes)

    elif volumes.shape[1] == 8:
        return hexa_to_quad(volumes)


def faces_to_edges(faces):
    """Compute edges based on following edge scheme.

    .. code-block::

        Ref: (node_ind), edge_ind

            (0)
             /\\
          0 /  \\2
           /____\\
        (1)  1   (2)

              2
        (3)*-----*(2)
           |     |
         3 |     | 1
        (0)*-----*(1)
              0

    Note: if `edges` index matter for tets, reorder it!

    Parameters
    -----------
    faces: (n, 3) or (n, 4) np.ndarray

    Returns
    --------
    edges: (n * 3, 2) or (n * 4, 2) np.ndarray
    """
    if faces.ndim != 2:
        raise ValueError(
            "Given input has wrong dimension. "
            "The input array for a faces has to be dim of 2"
        )

    num_faces = faces.shape[0]
    vertices_per_face = faces.shape[1]

    num_edges = int(num_faces * vertices_per_face)
    edges = np.ones((num_edges, 2), dtype=np.int32) * -1  # -1 for safety
    edges[:, 0] = faces.flatten()

    for i in range(vertices_per_face):
        # v_ind : corresponding vertex index for i value
        if i == int(vertices_per_face - 1):
            v_ind = 0
        else:
            v_ind = i + 1

        edges[i::vertices_per_face, 1] = faces[:, v_ind]

    # Quick sanity check - No entries are left untouched.
    if (edges == -1).any():
        raise ValueError("There was an error while computing edges.")

    return edges


def range_to_edges(range_, closed=False, continuous=True):
    """Given range, for example (a, b), returns an edge sequence that
    sequentially connects indices. If int is given as range, it is considered
    as (0, value). Used to be called "closed/open_loop_index_train".

    Parameters
    -----------
    range_: list, tuple, or int
    closed: bool
    continuous: bool

    Returns
    --------
    edges: (n, 2) np.ndarray
    """
    if isinstance(range_, int):
        indices = np.arange(range_)
    elif isinstance(range_, (list, tuple)):
        if len(range_) > 2:
            raise ValueError("Input range is too long")

        # keep the arange syntax. If this isn't wanted,
        # change to (ran[0], ran[1]+1)
        indices = np.arange(*range_)

    # closed is ignored
    if not continuous:
        if indices.size % 2 == 1:
            raise ValueError("Ranges should result in even number of indices.")
        return indices.reshape(-1, 2)

    # continuous edges
    indices = np.repeat(indices, 2)
    if closed:
        indices = np.append(indices, indices[0])[1:]
    else:
        indices = indices[1:-1]

    return indices.reshape(-1, 2)


def sequence_to_edges(seq, closed=False):
    """Given a sequence of int, "connect" to turn them into edges.

    Parameters
    -----------
    seq: (n,) array-like
    closed: bool

    Returns
    --------
    edges: (m, 2) np.ndarray
    """
    edges = np.repeat(seq, 2)[1:-1]
    edges = edges.reshape(-1, 2)
    if closed:
        edges = np.vstack((edges, [edges[-1, -1], edges[0, 0]]))

    return edges


def make_quad_faces(resolutions):
    """Given number of nodes per each dimension, returns connectivity
    information of a structured mesh. Counter clock wise connectivity.

    .. code-block::

        (3)*------*(2)
           |      |
           |      |
        (0)*------*(1)

    Parameters
    ----------
    resolutions: list

    Returns
    -------
    faces: (n, 4) np.ndarray
    """
    nnpd = np.asarray(resolutions)  # number of nodes per dimension
    if any(nnpd < 1):
        raise ValueError(f"The number of nodes per dimension is wrong: {nnpd}")

    total_nodes = np.product(nnpd)
    total_faces = (nnpd[0] - 1) * (nnpd[1] - 1)
    try:
        node_indices = np.arange(total_nodes).reshape(nnpd[1], nnpd[0])
    except ValueError as e:
        raise ValueError(f"Problem with generating node indices. {e}")

    faces = np.ones((total_faces, 4)) * -1

    faces[:, 0] = node_indices[: (nnpd[1] - 1), : (nnpd[0] - 1)].flatten()
    faces[:, 1] = node_indices[: (nnpd[1] - 1), 1 : nnpd[0]].flatten()
    faces[:, 2] = node_indices[1 : nnpd[1], 1 : nnpd[0]].flatten()
    faces[:, 3] = node_indices[1 : nnpd[1], : (nnpd[0] - 1)].flatten()

    if faces.all() == -1:
        raise ValueError("Something went wrong during `make_quad_faces`.")

    return faces.astype(np.int32)


def make_hexa_volumes(resolutions):
    """Given number of nodes per each dimension, returns connectivity
    information of structured hexahedron elements. Counter clock wise
    connectivity.

    .. code-block::

         (7)*-------*(6)
           /|      /|
          / | (5) / |
      (4)*-------*  |
         |  *----|--*(2)
         | /(3)  | /
         |/      |/
      (0)*-------*(1)

    Parameters
    -----------
    resolutions: list

    Returns
    --------
    elements: (n, 8) np.ndarray
    """
    nnpd = np.asarray(resolutions)  # number of nodes per dimension
    if any(nnpd < 1):
        raise ValueError(f"The number of nodes per dimension is wrong: {nnpd}")

    total_nodes = np.product(nnpd)
    total_volumes = np.product(nnpd - 1)
    node_indices = np.arange(total_nodes, dtype=np.int32).reshape(nnpd[::-1])

    volumes = np.ones((total_volumes, 8), dtype=np.int32) * int(-1)

    volumes[:, 0] = node_indices[
        : (nnpd[2] - 1), : (nnpd[1] - 1), : (nnpd[0] - 1)
    ].flatten()
    volumes[:, 1] = node_indices[
        : (nnpd[2] - 1), : (nnpd[1] - 1), 1 : nnpd[0]
    ].flatten()
    volumes[:, 2] = node_indices[
        : (nnpd[2] - 1), 1 : nnpd[1], 1 : nnpd[0]
    ].flatten()
    volumes[:, 3] = node_indices[
        : (nnpd[2] - 1), 1 : nnpd[1], : (nnpd[0] - 1)
    ].flatten()
    volumes[:, 4] = node_indices[
        1 : nnpd[2], : (nnpd[1] - 1), : (nnpd[0] - 1)
    ].flatten()
    volumes[:, 5] = node_indices[
        1 : nnpd[2], : (nnpd[1] - 1), 1 : nnpd[0]
    ].flatten()
    volumes[:, 6] = node_indices[
        1 : nnpd[2], 1 : nnpd[1], 1 : nnpd[0]
    ].flatten()
    volumes[:, 7] = node_indices[
        1 : nnpd[2], 1 : nnpd[1], : (nnpd[0] - 1)
    ].flatten()

    if (volumes == -1).any():
        raise ValueError("Something went wrong during `make_hexa_volumes`.")

    return volumes.astype(settings.INT_DTYPE)


def subdivide_edges(edges):
    """Subdivide edges. We assume that mid point is newly added points.

    ``Subdivided Edges``

    .. code-block::

        Edges (Lines)

        Ref: (node_ids), edge_ids

        (0)      (2)      (1)
         *--------*--------*

        edge_ids | node_ids
        ---------|----------
        0        | 0 2
        1        | 2 1

    Parameters
    -----------
    edges: (n, 2) np.ndarray

    Returns
    --------
    subdivided_edges: (n * 2, 2) np.ndarray
    """
    if edges.ndim != 2 or edges.shape[1] != 2:
        raise ValueError("Invalid edges shape!")

    raise NotImplementedError


def subdivide_tri(
    mesh,
    return_dict=False,
):
    """Subdivide triangles. Each triangle is divided into 4 meshes.

    ``Subdivided Faces``

    .. code-block::

        Triangles

        Ref: (node_ind), face_ind

                 (0)
                     _/|
                   _/ 0|
            (3)  _/____|(5)
               _/|    /|
             _/ 1| 3/ 2|
            /____|/____|
          (1)   (4)    (2)

        face_ind | node_ind
        ---------|----------
        0        | 0 3 5
        1        | 1 4 3
        2        | 2 5 4
        3        | 3 4 5

    Parameters
    -----------
    mesh: Mesh
    return_dict: bool

    Returns
    --------
    new_vertices: (n, d) np.ndarray
    subdivided_faces: (m, 3) np.ndarray
    mesh_dict: dict
      iff `return_dict=True`,
      returns dict(vertices=new_vertices, faces=subdivided_faces).
    """
    # This will only pass if the mesh is triangle mesh.
    if mesh.faces.shape[1] != 3:
        raise ValueError("Invalid faces shape!")

    # Form new vertices
    edge_mid_v = mesh.vertices[mesh.unique_edges().values].mean(axis=1)
    new_vertices = np.vstack((mesh.vertices, edge_mid_v))

    subdivided_faces = np.empty(
        (mesh.faces.shape[0] * 4, mesh.faces.shape[1]),
        dtype=np.int32,
    )

    mask = np.ones(subdivided_faces.shape[0], dtype=bool)
    mask[3::4] = False

    # 0th column minus (every 4th row, starting from 3rd row)
    subdivided_faces[mask, 0] = mesh.faces.flatten()  # copies

    # Form ids for new vertices
    new_vertices_ids = mesh.unique_edges().inverse + int(mesh.faces.max() + 1)
    # 1st & 2nd columns
    subdivided_faces[mask, 1] = new_vertices_ids
    subdivided_faces[mask, 2] = new_vertices_ids.reshape(-1, 3)[
        :, [2, 0, 1]
    ].flatten()

    # Every 4th row, starting from 3rd row
    subdivided_faces[~mask, :] = new_vertices_ids.reshape(-1, 3)

    if return_dict:
        return dict(
            vertices=new_vertices,
            faces=subdivided_faces,
        )

    else:
        return new_vertices, subdivided_faces


def subdivide_quad(
    mesh,
    return_dict=False,
):
    """Subdivide quads.

    Parameters
    -----------
    mesh: Mesh
    return_dict: bool

    Returns
    --------
    new_vertices: (n, d) np.ndarray
    subdivided_faces: (m, 4) np.ndarray
    mesh_dict: dict
      iff `return_dict=True`,
      returns dict(vertices=new_vertices, faces=subdivided_faces).
    """
    # This will only pass if the mesh is quad mesh.
    if mesh.faces.shape[1] != 4:
        raise ValueError("Invalid faces shape!")

    # Form new vertices
    edge_mid_v = mesh.vertices[mesh.unique_edges().values].mean(axis=1)
    face_centers = mesh.centers()
    new_vertices = np.vstack(
        (
            mesh.vertices,
            edge_mid_v,
            face_centers,
        )
    )

    subdivided_faces = np.empty(
        (mesh.faces.shape[0] * 4, mesh.faces.shape[1]),
        dtype=np.int32,
    )

    subdivided_faces[:, 0] = mesh.faces.flatten()
    subdivided_faces[:, 1] = mesh.unique_edges().inverse + len(mesh.vertices)
    subdivided_faces[:, 2] = np.repeat(
        np.arange(len(face_centers)) + (len(mesh.vertices) + len(edge_mid_v)),
        4,
        # dtype=np.int32,
    )
    subdivided_faces[:, 3] = (
        subdivided_faces[:, 1].reshape(-1, 4)[:, [3, 0, 1, 2]].flatten()
    )

    if return_dict:
        return dict(
            vertices=new_vertices,
            faces=subdivided_faces,
        )

    else:
        return new_vertices, subdivided_faces


def sorted_unique(connectivity, sorted_=False):
    """Given connectivity array, finds unique entries, based on its axis=1
    sorted values. Returned value will be sorted.

    Parameters
    -----------
    connectivity: (n, d) np.ndarray
    sorted_: bool

    Returns
    --------
    unique_info: Unique2DIntegers
    """
    s_connec = connectivity if sorted_ else np.sort(connectivity, axis=1)

    unique_stuff = arr.unique_rows(
        s_connec,
        return_index=True,
        return_inverse=True,
        return_counts=True,
        dtype_name=settings.INT_DTYPE,
    )

    return helpers.data.Unique2DIntegers(
        unique_stuff[0],  # values
        unique_stuff[1],  # ids
        unique_stuff[2],  # inverse
        unique_stuff[3],  # counts
    )



"""gustaf/gustaf/vertices.py.

Vertices. Base of all "Mesh" geometries.
"""

import copy

import numpy as np

from gustaf import helpers, settings, show, utils
from gustaf._base import GustafBase
from gustaf.helpers.options import Option


class VerticesShowOption(helpers.options.ShowOption):
    """
    Show options for vertices.
    """

    _valid_options = helpers.options.make_valid_options(
        *helpers.options.vedo_common_options,
        Option("vedo", "r", "Radius of vertices in units of pixels.", (int,)),
        Option(
            "vedo",
            "labels",
            "Places a label/description str at the place of vertices.",
            (np.ndarray, tuple, list),
        ),
        Option(
            "vedo",
            "label_options",
            "Label kwargs to be passed during initialization."
            "Valid keywords are: {scale: float, xrot: float, yrot: float, "
            "zrot: float, ratio: float, precision: int, italic: bool, "
            "font: str, justify: str, c: (str, tuple, list, int), "
            "alpha: float}. "
            "As further hint, justify takes '-' joined combination of "
            "{center, mid, right, left, top, bottom}.",
            (dict,),
        ),
    )

    _helps = "Vertices"

    def _initialize_vedo_showable(self):
        """
        Initialize Vertices showable for vedo.

        Parameters
        ----------
        None

        Returns
        -------
        vertices: vedo.Points
        """
        init_options = ("r",)

        vertices = show.vedo.Points(
            self._helpee.const_vertices, **self[init_options]
        )

        labels = self.get("labels", None)
        if labels is not None:
            # check length
            if len(labels) != len(self._helpee.const_vertices):
                raise ValueError(
                    f"number of label contents ({len(labels)}) and "
                    "number of vertices"
                    f"({len(self._helpee.const_vertices)}) does not match."
                )

            # apply options and return labels
            return vertices.labels(
                content=labels,
                on="points",
                **self.get("label_options", dict()),
            )

        else:
            # no labels, return Points
            return vertices


class Vertices(GustafBase):
    kind = "vertex"

    __slots__ = (
        "_vertices",
        "_const_vertices",
        "_computed",
        "_show_options",
        "_setter_copies",
        "_vertex_data",
    )

    # define frequently used types as dunder variable
    __show_option__ = VerticesShowOption

    def __init__(
        self,
        vertices=None,
        copy=True,
    ):
        """Vertices. It has vertices.

        Parameters
        -----------
        vertices: (n, d) np.ndarray
        copy: bool
          If false, all setter calls will try to avoid copy.

        Returns
        --------
        None
        """
        # call setters
        self.setter_copies = copy
        self.vertices = vertices

        # init helpers
        self._vertex_data = helpers.data.VertexData(self)
        self._computed = helpers.data.ComputedMeshData(self)
        self._show_options = self.__show_option__(self)

    @property
    def vertices(self):
        """Returns vertices.

        Parameters
        -----------
        None

        Returns
        --------
        vertices: (n, d) np.ndarray
        """
        self._logd("returning vertices")
        return self._vertices

    @vertices.setter
    def vertices(self, vs):
        """Vertices setter. This will saved as a tracked array. This tracked
        array is very sensitive and if we do anything with it that may hint an
        inplace operation, it will be marked as modified. This includes copying
        and slicing. If you know you aren't going to modify the array, please
        consider using `const_vertices`. Somewhat c-style hint in naming.

        Parameters
        -----------
        vs: (n, d) np.ndarray

        Returns
        --------
        None
        """
        self._logd("setting vertices")

        self._vertices = helpers.data.make_tracked_array(
            vs, settings.FLOAT_DTYPE, self.setter_copies
        )

        # shape check
        if self._vertices.size > 0:
            utils.arr.is_shape(self._vertices, (-1, -1), strict=True)

        # exact same, but not tracked.
        self._const_vertices = self._vertices.view()
        self._const_vertices.flags.writeable = False

        # at each setting, validate vertex_data
        # --> by len mismatch, will clear data
        if hasattr(self, "vertex_data"):
            self.vertex_data._validate_len(raise_=False)

    @property
    def const_vertices(self):
        """Returns non-mutable view of `vertices`. Naming inspired by c/cpp
        sessions.

        Parameters
        -----------
        None

        Returns
        --------
        None
        """
        self._logd("returning const_vertices")
        return self._const_vertices

    @property
    def vertex_data(self):
        """
        Returns vertex_data manager. Behaves similar to dict() and can be used
        to store values/data associated with each vertex.

        Parameters
        ----------
        None

        Returns
        -------
        vertex_data: VertexData
        """
        self._logd("returning vertex_data")
        return self._vertex_data

    @property
    def setter_copies(self):
        """
        Switch to set if setter should copy or try to avoid copying.
        If data is np.ndarray, c_contiguous, and has same dtype as
        corresponding settings.<FLOAT/INT>_dtype, it can probably avoid being
        copied. Setter will try to cast input to bool.

        Parameters
        ----------
        None

        Returns
        -------
        setter_copies: bool
        """
        return self._setter_copies

    @setter_copies.setter
    def setter_copies(self, should_copy):
        """
        Sets setter_copies

        Parameters
        ----------
        should_copy: bool

        Returns
        -------
        None
        """
        self._setter_copies = bool(should_copy)

    @property
    def show_options(self):
        """
        Returns a show option manager for this object. Behaves similar to
        dict.

        Parameters
        ----------
        None

        Returns
        -------
        show_options: ShowOption
          A derived class that's suitable for current class.
        """
        self._logd("returning show_options")
        return self._show_options

    @property
    def vis_dict(self):
        """
        Temporary backward compatibility
        """
        self._logw("`vis_dict` is deprecated. Please use `show_options`")
        return self.show_options

    @vis_dict.setter
    def vis_dict(self, vd):
        """
        Tmp
        """
        self._logw("`vis_dict` is deprecated. Please use `show_options`")
        self._show_options = vd

    @property
    def whatami(self):
        """Answers deep philosophical question: "what am i"?

        Parameters
        ----------
        None

        Returns
        --------
        whatami: str
          vertices
        """
        return "vertices"

    @helpers.data.ComputedMeshData.depends_on(["vertices"])
    def unique_vertices(self, tolerance=None, **kwargs):
        """Returns a namedtuple that holds unique vertices info. Unique here
        means "close-enough-within-tolerance".

        Parameters
        -----------
        tolerance: float
          (Optional) Default is settings.TOLERANCE
        recompute: bool
          Only applicable as keyword argument. Force re-computes.

        Returns
        --------
        unique_vertices_info: Unique2DFloats
          namedtuple with `values`, `ids`, `inverse`, `intersection`.
        """
        self._logd("computing unique vertices")
        if tolerance is None:
            tolerance = settings.TOLERANCE

        values, ids, inverse, intersection = utils.arr.close_rows(
            self.const_vertices, tolerance=tolerance
        )

        return helpers.data.Unique2DFloats(
            values,
            ids,
            inverse,
            intersection,
        )

    @helpers.data.ComputedMeshData.depends_on(["vertices"])
    def bounds(self):
        """Returns bounds of the vertices. Bounds means AABB of the geometry.

        Parameters
        -----------
        None

        Returns
        --------
        bounds: (d,) np.ndarray
        """
        self._logd("computing bounds")
        return utils.arr.bounds(self.const_vertices)

    @helpers.data.ComputedMeshData.depends_on(["vertices"])
    def bounds_diagonal(self):
        """Returns diagonal vector of the bounding box.

        Parameters
        -----------
        None

        Returns
        --------
        bounds_diagonal: (d,) np.ndarray
          same as `bounds[1] - bounds[0]`
        """
        self._logd("computing bounds_diagonal")
        bounds = self.bounds()
        return bounds[1] - bounds[0]

    @helpers.data.ComputedMeshData.depends_on(["vertices"])
    def bounds_diagonal_norm(self):
        """Returns norm of bounds diagonal.

        Parameters
        -----------
        None

        Returns
        --------
        bounds_diagonal_norm: float
        """
        self._logd("computing bounds_diagonal_norm")
        return float(sum(self.bounds_diagonal() ** 2) ** 0.5)

    def update_vertices(self, mask, inverse=None):
        """Update vertices with a mask. In other words, keeps only masked
        vertices. Adapted from `github.com/mikedh/trimesh`. Updates
        connectivity accordingly too.

        Parameters
        -----------
        mask: (n,) bool or int
        inverse: (len(self.vertices),) int

        Returns
        --------
        updated_self: type(self)
        """
        vertices = self.const_vertices.copy()

        # make mask numpy array
        mask = np.asarray(mask)

        if (mask.dtype.name == "bool" and mask.all()) or len(mask) == 0:
            return self

        # create inverse mask if not passed
        check_neg = False
        if inverse is None and self.kind != "vertex":
            inverse = np.full(len(vertices), -11, dtype=settings.INT_DTYPE)
            check_neg = True
            if mask.dtype.kind == "b":
                inverse[mask] = np.arange(mask.sum())
            elif mask.dtype.kind == "i":
                inverse[mask] = np.arange(len(mask))
            else:
                inverse = None

        # re-index elements from inverse
        # TODO: Here could be a good place to preserve BCs.
        elements = None
        if inverse is not None and self.kind != "vertex":
            elements = self.const_elements.copy()
            elements = inverse[elements.reshape(-1)].reshape(
                (-1, elements.shape[1])
            )
            # remove all the elements that's not part of inverse
            if check_neg:
                elem_mask = (elements > -1).all(axis=1)
                elements = elements[elem_mask]

        # apply mask
        vertices = vertices[mask]

        def update_vertex_data(obj, m, vertex_data=None):
            """apply mask to vertex data if there's any."""
            new_data = helpers.data.VertexData(obj)
            if vertex_data is None:
                vertex_data = obj.vertex_data

            for key, values in vertex_data.items():
                # should work, since this is called after updating vertices
                new_data[key] = values[m]

            obj._vertex_data = new_data

            return obj

        # update
        self.vertices = vertices
        if elements is not None:
            self.elements = elements

        update_vertex_data(self, mask)

        return self

    def select_vertices(self, ranges):
        """Returns vertices inside the given range.

        Parameters
        -----------
        ranges: (d, 2) array-like
          Takes None.

        Returns
        --------
        ids: (n,) np.ndarray
        """
        return utils.arr.select_with_ranges(self.vertices, ranges)

    def remove_vertices(self, ids):
        """Removes vertices with given vertex ids.

        Parameters
        -----------
        ids: (n,) np.ndarray

        Returns
        --------
        new_self: type(self)
        """
        mask = np.ones(len(self.vertices), dtype=bool)
        mask[ids] = False

        return self.update_vertices(mask)

    def merge_vertices(self, tolerance=None):
        """Based on unique vertices, merge vertices if it is mergeable.

        Parameters
        -----------
        tolerance: float
          Default is settings.TOLERANCE

        Returns
        --------
        merged_self: type(self)
        """
        unique_vs = self.unique_vertices()

        self._logd("number of vertices")
        self._logd(f"  before merge: {len(self.vertices)}")
        self._logd(f"  after merge: {len(unique_vs.ids)}")

        return self.update_vertices(
            mask=unique_vs.ids,
            inverse=unique_vs.inverse,
        )

    def showable(self, **kwargs):
        """Returns showable object, meaning object of visualization backend.

        Parameters
        -----------
        **kwargs:

        Returns
        --------
        showable: obj
          Obj of `gustaf.settings.VISUALIZATION_BACKEND`
        """
        return show.make_showable(self, **kwargs)

    def show(self, **kwargs):
        """Show current object using visualization backend.

        Parameters
        -----------
        **kwargs:


        Returns
        --------
        None
        """
        return show.show(self, **kwargs)

    def copy(self):
        """Returns deepcopy of self.

        Parameters
        -----------
        None

        Returns
        --------
        self_copy: type(self)
        """
        # all attributes are deepcopy-able
        return copy.deepcopy(self)

    @classmethod
    def concat(cls, *instances):
        """Sequentially put them together to make one object.

        Parameters
        -----------
        *instances: List[type(cls)]
          Allows one iterable object also.

        Returns
        --------
        one_instance: type(cls)
        """

        def is_concatable(inst):
            """Return true, if it is same as type(cls)"""
            if isinstance(inst, cls):
                return True
            else:
                return False

        # If only one instance is given and it is iterable, adjust
        # so that we will just iterate that.
        if (
            len(instances) == 1
            and not isinstance(instances[0], str)
            and hasattr(instances[0], "__iter__")
        ):
            instances = instances[0]

        vertices = []
        has_elem = cls.kind != "vertex"
        if has_elem:
            elements = []

        # check if everything is "concatable".
        for ins in instances:
            if not is_concatable(ins):
                raise TypeError(
                    "Can't concat. One of the instances is not "
                    f"`{cls.__name__}`."
                )

            # make sure each element index starts from 0 & end at len(vertices)
            tmp_ins = ins.copy().remove_unreferenced_vertices()

            vertices.append(tmp_ins.vertices.copy())

            if has_elem:
                if len(elements) == 0:
                    elements.append(tmp_ins.elements.copy())
                    e_offset = elements[-1].max() + 1

                else:
                    elements.append(
                        # copy is not necessary here,
                        tmp_ins.elements
                        + e_offset
                    )
                    e_offset = elements[-1].max() + 1

        if has_elem:
            return cls(
                vertices=np.vstack(vertices),
                elements=np.vstack(elements),
            )

        else:
            return Vertices(vertices=np.vstack(vertices))

    def __add__(self, to_add):
        """Concat in form of +.

        Parameters
        -----------
        to_add: type(self)

        Returns
        --------
        added: type(self)
        """
        return type(self).concat(self, to_add)
